
/*
   Smart Thermostat by Roque Campos
   Sketch to read temperatures with an ESP32 board and a DHTxx sensor connected to the pin 26, and to control a Relay connected to the Pin 27
   Complete Project Details https://www.smart-thermostat.eu
*/

#include <WiFi.h>
#include <DHTesp.h>
#include "esp_system.h"
#include "soc/rtc.h"
#include "rom/uart.h"

/****** Start configuration section ******/

const char* ssid     = "YourWifiName"; //Replace with your network SSID
const char* password = "YourPassword"; //Replace with your network password
const String nodeName = "NodeName"; //Replace with the name for this node (example "Beed Room")


//If you are using a HIGH LEVEL TRIGGERED Relay set the following flag to true
//If you are using a LOW LEVEL TRIGGERED Relay set the following flag to false
bool RELAY_IS_HIGH_LEVEL = true;

//Uncomment the appropiate type for your DHT Sensor.
DHTesp::DHT_MODEL_t DHT_TYPE = DHTesp::DHT22;
//DHTesp::DHT_MODEL_t DHT_TYPE=DHTesp::DHT11;

/****** End configuration section ******/
/****** Please don't touch anything below this point unless you know what you are doing. ******/

const String sketchVersion = "4";
const String nodeType = "TH&Relay";
unsigned long lastReceivedRelayCommandMilis;
const unsigned int TIMEOUT_REICEIVED_COMMANDS_MILIS = 1000 * 60 * 2; //Two minutes
String relayError = "";
String thError = "";
WiFiServer server(80);
const int RELAY_PIN = 27;
const int ONBOARD_LED_PIN = 2;
const int DHT_PIN = 26;
DHTesp dht;



void setup() {
  Serial.begin(115200);
  slowDownCpu();
  pinMode(ONBOARD_LED_PIN, OUTPUT);

  pinMode(RELAY_PIN, OUTPUT);
  turnOffBoiler();
  dht.setup(DHT_PIN, DHT_TYPE);

  setUpWifi();

  lastReceivedRelayCommandMilis = millis();

}

void slowDownCpu() {
  rtc_clk_cpu_freq_set(RTC_CPU_FREQ_160M);
  uart_tx_wait_idle(0);
  int clockspeed = rtc_clk_cpu_freq_get();

  char* clockSpeeds[5] = {"XTAL", "80Mhz", "160Mhz", "240Mhz", "2Mhz"};
  Serial.print("Setting CPU freq to: ");
  Serial.println(clockSpeeds[clockspeed]);
  Serial.println("");
}

void setUpWifi() {
  //WiFi.config(ip,dns,gateway,subnet);


  Serial.println();
  Serial.println();
  Serial.print("Connecting to ");
  Serial.println(ssid);


  // attempt to connect to Wifi network:
  int counter = 0;
  while (WiFi.status() != WL_CONNECTED) {
    WiFi.disconnect();
    WiFi.begin(ssid, password);
    int counter = 0;
    while (WiFi.status() != WL_CONNECTED) {
      delay(1000);
      Serial.print(".");
      counter++;
      if (counter > 10) {
        break;
      }
    }
  }

  Serial.println("");
  Serial.println("WiFi connected");
  Serial.print("IP address: ");
  Serial.println(WiFi.localIP());
  Serial.print("MAC: ");
  Serial.println(WiFi.macAddress());
  // print the received signal strength:
  long rssi = WiFi.RSSI();
  Serial.print("Signal strength (RSSI):");
  Serial.print(rssi);
  Serial.println(" dBm");

  server.begin();

}

double celsius2Fahrenheit(double celsius)
{
  return celsius * 9 / 5 + 32;
}

void turnBoilerOffIfTimeout() {
  unsigned long now = millis();
  long elapsed = now - lastReceivedRelayCommandMilis;
  if (elapsed > TIMEOUT_REICEIVED_COMMANDS_MILIS) {
    turnOffBoiler();
    relayError = "Error: thermostat command timeout...";
  }
}

void turnOffBoiler() {
  if (RELAY_IS_HIGH_LEVEL) {
    digitalWrite(RELAY_PIN, LOW);
  } else {
    digitalWrite(RELAY_PIN, HIGH);
  }
}

void turnOnBoiler() {
  if (RELAY_IS_HIGH_LEVEL) {
    digitalWrite(RELAY_PIN, HIGH);
  } else {
    digitalWrite(RELAY_PIN, LOW);
  }
}

void loop() {

  if (WiFi.status() != WL_CONNECTED) {
    turnOffBoiler();
    relayError = "Error: wifi connection lost...";
    setUpWifi();
  } else {
    turnBoilerOffIfTimeout();
    // listen for incoming clients
    WiFiClient client = server.available();
    if (client.connected()) {
      digitalWrite(ONBOARD_LED_PIN, HIGH);

      String req = client.readStringUntil('\r');
      Serial.println("Received request: " + req);

      if (req.startsWith("GET /L")) {
        Serial.println("Relay Low");
        lastReceivedRelayCommandMilis = millis();
        turnOffBoiler();
        relayError = "";
      } else if (req.startsWith("GET /H")) {
        Serial.println("Relay High");
        lastReceivedRelayCommandMilis = millis();
        turnOnBoiler();
        relayError = "";
      }


      double humidity = dht.getHumidity();
      double celsiusTemp = dht.getTemperature();
      double fahrenheitTemp = -10000;
      int currentStatus = digitalRead(RELAY_PIN);


      if (dht.getStatus() != 0) {
        thError = String(dht.getStatusString());
        Serial.println("DHT Error status: " + thError);
        humidity = -10000;
        celsiusTemp = -10000;
      } else {
        thError = "";
        fahrenheitTemp = celsius2Fahrenheit(celsiusTemp);
        Serial.println("DHT OK");
      }

      // send a standard http response header
      String response = String();
      response += "HTTP/1.1 200 OK\n";
      response += "Content-Type: application/json\n";
      response += "\n"; //This carry return is very important. Without it the response will not be sent.
      response += "{\n";
      response += "\n";
      response += "\"type\":\"" + nodeType + "\",\n";
      response += "\"name\":\"" + nodeName + "\",\n";
      response += "\"version\":\"" + sketchVersion + "\",\n";
      response += "\"relay_error\":\"" + relayError + "\",\n";
      response += "\"th_error\":\"" + thError + "\",\n";
      response += "\"mac\":\"" + WiFi.macAddress() + "\",\n";
      response += "\"relay_high\": \"" + String(currentStatus) + "\",\n";
      response += "\"relay_high_level\": \"" + String(RELAY_IS_HIGH_LEVEL) + "\",\n";

      String cc = String(celsiusTemp);
      cc.trim();
      response += "\"celsius\":\"" + cc + "\",\n";

      String ff = String(fahrenheitTemp);
      ff.trim();
      response += "\"fahrenheit\":\"" + ff + "\",\n";

      String hh = String(humidity);
      hh.trim();
      response += "\"humidity\":\"" + hh + "\",\n";

      long rssi = WiFi.RSSI();
      response += "\"wifi_ssi\": \"" + String(rssi) + "\"\n";
      response += "}\n";
      response += "\n";
      Serial.println(response);
      client.println(response);
      delay(100);
      client.stop();
      digitalWrite(ONBOARD_LED_PIN, LOW);
      Serial.println("_________________________________________");
    }

  }
}
