/*
   Smart Thermostat by Roque Campos
   Sketch to siwtch on/off a relay with an ESP32
   Complete Project Details https://www.smart-pi-thermostat.com
*/

#include <WiFi.h>
#include "esp_system.h"
#include "soc/rtc.h"
#include "rom/uart.h"



/****** Start configuration section ******/

const char* ssid     = "YourWifiName"; //Replace with your network SSID
const char* password = "YourPassword"; //Replace with your network password
const String nodeName = "NodeName"; //Replace with the name for this node (example "Beed Room")

//If you are using a HIGH LEVEL TRIGGERED Relay set the following flag to true
//If you are using a LOW LEVEL TRIGGERED Relay set the following flag to false
bool RELAY_IS_HIGH_LEVEL = true;

/****** End configuration section ******/
/****** Please don't touch anything below this point unless you know what you are doing. ******/

const String sketchVersion = "4";
const String nodeType = "Relay";
unsigned long lastReceivedRelayCommandMilis;
const unsigned int TIMEOUT_REICEIVED_COMMANDS_MILIS = 1000 * 60 * 2; //Two minutes
String relayError = "";

const int RELAY_PIN = 27;
const int ONBOARD_LED_PIN = 2;

WiFiServer server(80);

void setup()
{
  Serial.begin(115200);
  slowDownCpu();
  pinMode(ONBOARD_LED_PIN, OUTPUT);
  pinMode(RELAY_PIN, OUTPUT);
  turnOffBoiler();
  delay(10);
  setUpWifi();
  lastReceivedRelayCommandMilis = millis();

}

void slowDownCpu() {
  rtc_clk_cpu_freq_set(RTC_CPU_FREQ_160M);
  uart_tx_wait_idle(0);
  int clockspeed = rtc_clk_cpu_freq_get();

  char* clockSpeeds[5] = {"XTAL", "80Mhz", "160Mhz", "240Mhz", "2Mhz"};
  Serial.print("Setting CPU freq to: ");
  Serial.println(clockSpeeds[clockspeed]);
  Serial.println("");
}

void setUpWifi() {
  // attempt to connect to Wifi network:
  int counter = 0;
  while (WiFi.status() != WL_CONNECTED) {
    WiFi.disconnect();
    WiFi.begin(ssid, password);
    int counter = 0;
    while (WiFi.status() != WL_CONNECTED) {
      delay(1000);
      Serial.print(".");
      counter++;
      if (counter > 10) {
        break;
      }
    }
  }

  Serial.println("");
  Serial.println("WiFi connected");
  Serial.print("IP address: ");
  Serial.println(WiFi.localIP());
  Serial.print("MAC: ");
  Serial.println(WiFi.macAddress());
  // print the received signal strength:
  long rssi = WiFi.RSSI();
  Serial.print("Signal strength (RSSI):");
  Serial.print(rssi);
  Serial.println(" dBm");
  server.begin();

}

void turnBoilerOffIfTimeout() {
  unsigned long now = millis();
  long elapsed = now - lastReceivedRelayCommandMilis;
  if (elapsed > TIMEOUT_REICEIVED_COMMANDS_MILIS) {
    turnOffBoiler();
    relayError = "Error: thermostat command timeout...";
  }
}

void turnOffBoiler() {
  if (RELAY_IS_HIGH_LEVEL) {
    digitalWrite(RELAY_PIN, LOW);
  } else {
    digitalWrite(RELAY_PIN, HIGH);
  }
}

void turnOnBoiler() {
  if (RELAY_IS_HIGH_LEVEL) {
    digitalWrite(RELAY_PIN, HIGH);
  } else {
    digitalWrite(RELAY_PIN, LOW);
  }
}


void loop() {

  if (WiFi.status() != WL_CONNECTED) {
    turnOffBoiler();
    relayError = "Error: wifi connection lost...";
    setUpWifi();
  } else {
    turnBoilerOffIfTimeout();
    WiFiClient client = server.available();   // listen for incoming clients

    if (client.connected()) {                             // if you get a client,
      digitalWrite(ONBOARD_LED_PIN, HIGH);
      Serial.println("New Client.");           // print a message out the serial port

      String req = client.readStringUntil('\r');
      Serial.println("Received request: " + req);

      if (req.startsWith("GET /L")) {
        Serial.println("Relay Low");
        lastReceivedRelayCommandMilis = millis();
        turnOffBoiler();
        relayError = "";
      } else if (req.startsWith("GET /H")) {
        Serial.println("Relay High");
        lastReceivedRelayCommandMilis = millis();
        turnOnBoiler();
        relayError = "";
      }


      // send a standard http response header
      String response = String();
      response += "HTTP/1.1 200 OK\n";
      response += "Content-Type: application/json\n";
      response += "\n"; //This carry return is very important. Without it the response will not be sent.
      response += "{\n";
      response += "\n";
      response += "\"type\":\"" + nodeType + "\",\n";
      response += "\"name\":\"" + nodeName + "\",\n";
      response += "\"version\":\"" + sketchVersion + "\",\n";
      response += "\"relay_error\":\"" + relayError + "\",\n";
      response += "\"mac\":\"" + WiFi.macAddress() + "\",\n";

      int currentStatus = digitalRead(RELAY_PIN);
      response += "\"relay_high\": \"" + String(currentStatus) + "\",\n";
      response += "\"relay_high_level\": \"" + String(RELAY_IS_HIGH_LEVEL) + "\",\n";

      long rssi = WiFi.RSSI();
      response += "\"wifi_ssi\": \"" + String(rssi) + "\"\n";
      response += "}\n";
      response += "\n";
      Serial.println(response);
      client.println(response);
      delay(100);
      client.stop();
      digitalWrite(ONBOARD_LED_PIN, LOW);
      Serial.println("_________________________________________");
    }

  }
}
